/*
 * k-nearest neighbors with squared Mahalanobis distance metric
 * MEMOCODE 2014 Design Contest
 * Peter Milder, peter.milder@stonybrook.edu
 *
 * For more information, please see README.TXT and
 * http://www.ece.stonybrook.edu/~pmilder/memocode/
 *
 * This program provides the functional reference software implementation.
 *
 * Usage: ./kNN setupfile
 * See README.TXT for input/output specification, or run: 
 *     make runsmall
 * to compile and run the small example. 
 */

/*
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */


#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <limits.h>
#include <assert.h>
#include <string.h>
#include <math.h>

typedef long int TYPE;

int K;
int D;
int DATASETSIZE;
int INPUTSETSIZE;
char benchmarkName[256];

// Function prototypes
void initValues(TYPE **dataset, TYPE **inputset, TYPE **inverseS, TYPE **lowestDistances, int **lowestDistIdx);
TYPE mahalanobisDistanceSq(TYPE *x, TYPE *y, TYPE **inverseS);
void findKLowest(TYPE *lowestDist, int *lowestIdx, TYPE newDist, int newIdx);
void outputAndCleanup(TYPE **dataset, TYPE **inputset, TYPE **inverseS, TYPE **lowestDistances, int **lowestDistIdx);
void getSettings();


int main(int argc, char *argv[]) {
	int l, i, j;
	
	if (argc != 2) {
		printf("Usage: %s setupfile\n", argv[0]);
	}
	else {
		getSettings(argv);
	
		///////////////////////////////////////////////////////////
		// Declare variables and call initValues to initialize them.
		// Inputs: 
		//   dataset:  set of known vectors with which we want to compare each input vector
		//   inputset: set of input vectors
		//   inverseS: inverse covariance matrix
		// Outputs:
		//   lowestDistances: the K lowest distances for each input vector
		//   lowestDistIdx:   the associated indices into dataset

		// Declare variables and malloc pointers
		TYPE **dataset         = malloc(sizeof(*dataset) * DATASETSIZE);
		TYPE **inputset        = malloc(sizeof(*inputset) * INPUTSETSIZE);
		TYPE **inverseS        = malloc(sizeof(*inverseS) * D);
		TYPE **lowestDistances = malloc(sizeof(*lowestDistances) * INPUTSETSIZE);		
		int  **lowestDistIdx   = malloc(sizeof(*lowestDistIdx) * INPUTSETSIZE);
		
		// Initialize and set input values from files 
		initValues(dataset, inputset, inverseS, lowestDistances, lowestDistIdx);

		/////////////////////////////////////////////////////////			
		// begin timing here

		for (i=0; i<INPUTSETSIZE; i++) {		
			for (j=0; j<DATASETSIZE; j++) {
				TYPE thisDist = mahalanobisDistanceSq(&dataset[j][0], &inputset[i][0], inverseS);							
				findKLowest(&lowestDistances[i][0], &lowestDistIdx[i][0], thisDist, j);
			}				
		}

		// end timing here
		/////////////////////////////////////////////////////////			

		// print the output values; free allocated memory
		outputAndCleanup(dataset, inputset, inverseS, lowestDistances, lowestDistIdx);
	} 
	
	return 0;
}



void getSettings(char* argv[]) {
	int z;
	FILE *f;
	if ((f = fopen(argv[1], "r")) == NULL) {
		printf("Cannot open settings file %s\n", argv[1]);
		assert(0);
	}
	printf("Reading settings from %s.\n", argv[1]);
	z = fscanf(f, "%d", &K);	
	z = fscanf(f, "%d", &D);	
	z = fscanf(f, "%d", &DATASETSIZE);	
	z = fscanf(f, "%d", &INPUTSETSIZE);	
	z = fscanf(f, "%s", benchmarkName);
	fclose(f);
}


// Allocate memory and read data from input files
void initValues(TYPE **dataset, TYPE **inputset, TYPE **inverseS, TYPE **lowestDistances, int **lowestDistIdx) {
	int i, j, z;

	//////////////////////////////////////////////////////
	// Allocate memory
	if (dataset) {
		for (i=0; i<DATASETSIZE; i++) {
			dataset[i] = malloc(sizeof(TYPE) * D);
		}
	}
	if (inverseS) {
		for (i=0; i<D; i++)
			inverseS[i] = malloc(sizeof(TYPE) * D);
	}
	if (inputset) {
		for (i=0; i<INPUTSETSIZE; i++)
			inputset[i] = malloc(sizeof(TYPE) * D);
	}
	if (lowestDistances) {
		for (i=0; i<INPUTSETSIZE; i++)
			lowestDistances[i] = malloc(sizeof(TYPE) * K);
	}
	
	if (lowestDistIdx) {
		for (i=0; i<INPUTSETSIZE; i++)
			lowestDistIdx[i] = malloc(sizeof(int) * K);
	}
	
	
	//////////////////////////////////////////////////////
	// Read values from input files.

	// Read the values of dataset from dataset.txt
	char buf[256];
	snprintf(buf, sizeof buf, "%s.dataset", benchmarkName);
	FILE *f;
	if ((f = fopen(buf, "r")) == NULL ) {
        printf("Can't open file %s\n", buf);
        assert(0);
    } 
    
    for (i=0; i < DATASETSIZE; i++) {
    	for (j=0; j < D; j++) {
    		int x;
    		z = fscanf(f, "%d", &x);
    		dataset[i][j] = (TYPE)x;
    	}
    }
    fclose(f);
		
	// Initialize the lowest distances seen to the maximum representable value
	for (i=0; i<INPUTSETSIZE; i++) 
		for (j=0; j<K; j++) 
			lowestDistances[i][j] = LONG_MAX; 

	// Read the input values from inputs.txt
	snprintf(buf, sizeof buf, "%s.input", benchmarkName);
	if ((f = fopen(buf, "r" )) == NULL ) {
        printf("Can't open file %s\n", buf);
        assert(0);
    } 
    
    for (i=0; i < INPUTSETSIZE; i++) {
    	for (j=0; j < D; j++) {
    		int x;
    		z = fscanf(f, "%d", &x);
    		inputset[i][j] = (TYPE)x;
    	}
    }
    fclose(f);
    
    // Read the inv. Covariance matrix values 
   	snprintf(buf, sizeof buf, "%s.invS", benchmarkName);
	if ((f = fopen(buf, "r")) == NULL ) {
        printf("Can't open file %s\n", buf);
        assert(0);
    } 
    
    for (i=0; i < D; i++) {
    	for (j=0; j < D; j++) {
    		int x;
    		z = fscanf(f, "%d", &x);
    		inverseS[i][j] = (TYPE)x;
    	}
    }
    fclose(f);    
 
}

// Calculate the squared Mahalanobis Distance between x and y, with inverseS as the 
// inverse covariance matrix.
TYPE mahalanobisDistanceSq(TYPE *x, TYPE *y, TYPE **inverseS) {
	int i, j;

	TYPE dif[D];
	
	// Calculate difference: x-y
	for (i=0; i<D; i++) {
		dif[i] = x[i]-y[i];
	}

	// multiply matrix inverseS times column vector dif
	TYPE mvmRes[D];	
	for (i=0; i<D; i++) {
		mvmRes[i] = 0;
		for (j=0; j<D; j++) {
			mvmRes[i] += inverseS[i][j] * dif[j];
		}		
	}
	
	// multiply dif (as a row vector) times mvmRes (column vector)
	TYPE distance = 0;
	for (i=0; i<D; i++) {
		distance += dif[i] * mvmRes[i];
	}
	
	return distance;
}

// Maintains the list of K lowest distances seen and their indices
//   lowestDist: list of K lowest distances, sorted from low to high
//   lowestIdx:  their indices
//   newDist:    distance for this iteration
//   newIdx:     iteration index
void findKLowest(TYPE *lowestDist, int *lowestIdx, TYPE newDist, int newIdx) {
	int i;
	
	// Where do we insert this new value? 
	int replaceIdx = -1;
	for (i=0; i<K; i++) {
		if (newDist < lowestDist[i]) {
			replaceIdx = i;
			break;
		}
	}
	
	// If we are inserting, move the larger to the right by 1 and insert this one
	if (replaceIdx >= 0) {
		for (i=K-1; i>replaceIdx; i--) {
			lowestDist[i] = lowestDist[i-1];
			lowestIdx[i] = lowestIdx[i-1];			
		}
		lowestDist[replaceIdx] = newDist;
		lowestIdx[replaceIdx] = newIdx;
	}
}

// Writes the output files and frees allocated memory
void outputAndCleanup(TYPE **dataset, TYPE **inputset, TYPE **inverseS, TYPE **lowestDistances, int **lowestDistIdx) {
	int i, l;
	FILE *f;

	char buf[256];
	snprintf(buf, sizeof buf, "%s.dists", benchmarkName);
	if ((f = fopen(buf, "w")) == NULL) {
        printf("Can't open file %s\n", buf);
        assert(0);
    } 

	for (i=0; i<INPUTSETSIZE; i++) {
		for (l=0; l<K; l++) {
			fprintf(f, "%ld ", lowestDistances[i][l]);
		}
		fprintf(f, "\n");
	}
	fclose(f);

	snprintf(buf, sizeof buf, "%s.indices", benchmarkName);
	if ((f = fopen(buf, "w")) == NULL) {
        printf("Can't open file %s\n", buf);
        assert(0);
    } 

	for (i=0; i<INPUTSETSIZE; i++) {
		for (l=0; l<K; l++) {
			fprintf(f, "%d ", lowestDistIdx[i][l]);
		}
		fprintf(f, "\n");
	}
	fclose(f);

	// free	
	for (i=0; i<DATASETSIZE; i++) 
		free(dataset[i]);
	free(dataset);
	
	for (i=0; i<INPUTSETSIZE; i++) 
		free(inputset[i]);
	free(inputset);
	
	for (i=0; i<D; i++) 
		free(inverseS[i]);
	free(inverseS);
	
	for (i=0; i<INPUTSETSIZE; i++) 
		free(lowestDistances[i]);
	free(lowestDistances);
	
	for (i=0; i<INPUTSETSIZE; i++) 
		free(lowestDistIdx[i]);
	free(lowestDistIdx);
}


